# --------------------------------------------------------------------------
# Default template loaders.
# --------------------------------------------------------------------------

import os

from .templates import Template
from .errors import LoadError, TemplateNotFound


class FileLoader:

    """ Loads templates from the file system. Assumes files are utf-8 encoded.

    Compiled templates are cached in memory, so they only need to be
    compiled once. Templates are *not* automatically recompiled if the
    underlying template file changes.

    A FileLoader instance should be initialized with a path to a base
    template directory.

        loader = FileLoader('/path/to/base/directory')

    The loader instance can then be called with one or more path strings.
    The loader will return the template object corresponding to the first
    existing template file or raise a TemplateNotFound exception if no file
    can be located.

    Note that the path strings may include subdirectory paths:

        template = loader('foo.txt')
        template = loader('subdir/foo.txt')

    """

    def __init__(self, root):
        self.root = root
        self.cache = {}

    def __call__(self, *paths):
        for path in paths:

            if path in self.cache:
                return self.cache[path]

            fullpath = os.path.join(self.root, path)
            if os.path.isfile(fullpath):
                try:
                    with open(fullpath, encoding='utf-8') as file:
                        return self.cache.setdefault(path, Template(file.read()))
                except OSError:
                    raise LoadError("error loading template file [%s]" % path)

        raise TemplateNotFound()


class FileReloader:

    """ Loads templates from the file system. Assumes files are utf-8 encoded.

    Compiled templates are cached in memory, so they only need to be
    compiled once. Templates are automatically recompiled if the
    underlying template file changes.

    """

    def __init__(self, root):
        self.root = root
        self.cache = {}

    def __call__(self, *paths):
        for path in paths:

            fullpath = os.path.join(self.root, path)
            if os.path.isfile(fullpath):
                try:
                    mtime = os.path.getmtime(fullpath)
                    if path in self.cache:
                        if mtime == self.cache[path][0]:
                            return self.cache[path][1]
                    with open(fullpath, encoding='utf-8') as file:
                        self.cache[path] = (mtime, Template(file.read()))
                        return self.cache[path][1]
                except OSError:
                    raise LoadError("error loading template file [%s]" % path)

        raise TemplateNotFound()


class DictLoader:

    """ Loads templates from a dictionary of template strings. """

    def __init__(self, strings):
        self.templates = {}
        self.strings = strings

    def __call__(self, *names):
        for name in names:
            if name in self.templates:
                return self.templates[name]
            elif name in self.strings:
                return self.templates.setdefault(name, Template(self.strings[name]))
        raise TemplateNotFound()



# --------------------------------------------------------------------------
# Ibis: a lightweight template engine.
#
# How it works: A lexer transforms a template string into an iterable
# sequence of tokens. A parser takes this sequence and compiles it into a
# tree of nodes. Each node has a .render() method which takes a context
# object and returns a string. The entire compiled node tree can be rendered
# by calling .render() on the root node.
#
# Compiling and rendering the node tree are two distinct processes. Once
# the template has been compiled it can be cached and rendered multiple times
# with different context objects.
#
# The Template class acts as the public interface to the template engine.
# This is the only class the end-user needs to interact with directly.
#
# A Template object is initialized with a template string. It compiles the
# string and stores the resulting node tree for future rendering. Calling the
# template object's .render() method with a dictionary of key-value pairs or
# a set of keyword arguments renders the template and returns the result as a
# string.
#
# Example:
#
#     >>> template = Template('{{foo}} and {{bar}}')
#
#     >>> template.render(foo='ham', bar='eggs')
#     'ham and eggs'
#
#     >>> template.render({'foo': 1, 'bar': 2})
#     '1 and 2'
#
# Author: Darren Mulholland <darren@mulholland.xyz>
# License: Public Domain
# --------------------------------------------------------------------------


# Library version number.
__version__ = "1.6.0"


# Import modules to make them available to callers via a simple 'import ibis'
# statement. Otherwise callers would have to 'import.foo' for each
# individual module.
from . import config
from . import filters
from . import nodes
from . import loaders
from . import errors

from .templates import Template



# --------------------------------------------------------------------------
# Default filter functions for use in templates. Additional filter functions
# can be registered using the `@register` decorator:
#
#     @ibis.filters.register('name')
#
# A filter function should accept at least one argument - the value to be
# filtered - and return the filtered result. It can optionally accept any
# number of additional arguments.
# --------------------------------------------------------------------------

import random
import re
import pprint
import html

from . import errors

try:
    import pygments
    import pygments.lexers
    import pygments.formatters
except ImportError:
    pygments = None


# Dictionary of registered filter functions.
filtermap = {}


def register(nameorfunc=None):

    """ Decorator function for registering filters.

    Can be used as:

        @register
        @register()
        @register('name')

    If no name is supplied the function name will be used.

    """

    if callable(nameorfunc):
        filtermap[nameorfunc.__name__] = nameorfunc
        return nameorfunc

    def register_filter_function(func):
        filtermap[nameorfunc or func.__name__] = func
        return func

    return register_filter_function


@register
def argtest(*args):
    """ Test filter: returns arguments as a concatenated string. """
    return '|'.join(str(arg) for arg in args)


@register
def default(obj, fallback):
    """ Returns `obj` if `obj` is truthy, otherwise `fallback`. """
    return obj or fallback


@register
def dtformat(dt, format='%Y-%m-%d %H:%M'):
    """ Formats a datetime object using the specified format string. """
    return dt.strftime(format)


@register
def endswith(s, suffix):
    """ True if the string ends with the specified suffix. """
    return s.endswith(suffix)


@register
@register('e')
@register('esc')
def escape(s, quotes=True):
    """ Converts html syntax characters to character entities. """
    return html.escape(s, quotes)


@register
def first(seq):
    """ Returns the first element in the sequence `seq`. """
    return seq[0]


@register
def firsth(html):
    """ Returns the content of the first heading element. """
    match = re.search(r'<h(\d)+[^>]*>(.*?)</h\1>', html, flags=re.DOTALL)
    return match.group(2) if match else ''


@register
def firsth1(html):
    """ Returns the content of the first h1 element. """
    match = re.search(r'<h1[^>]*>(.*?)</h1>', html, flags=re.DOTALL)
    return match.group(1) if match else ''


@register
def firstp(html):
    """ Returns the content of the first p element. """
    match = re.search(r'<p[^>]*>(.*?)</p>', html, flags=re.DOTALL)
    return match.group(1) if match else ''


@register('reversed')
def get_reversed(seq):
    """ Returns a reverse iterator over the sequence `seq`. """
    return reversed(seq)


@register
def index(seq, i):
    """ Returns the ith element in the sequence `seq`. """
    return seq[i]


@register('divisible_by')
def is_divisible_by(n, d):
    """ True if the integer `n` is a multiple of the integer `d`. """
    return n % d == 0


@register('even')
def is_even(n):
    """ True if the integer `n` is even. """
    return n % 2 == 0


@register('odd')
def is_odd(n):
    """ True if the integer `n` is odd. """
    return n % 2 != 0


@register
def join(seq, sep=''):
    """ Joins elements of the sequence `seq` with the string `sep`. """
    return sep.join(str(item) for item in seq)


@register
def last(seq):
    """ Returns the last element in the sequence `seq`. """
    return seq[-1]


@register('len')
def length(seq):
    """ Returns the length of the sequence `seq`. """
    return len(seq)


@register
def lower(s):
    """ Returns the string `s` converted to lowercase. """
    return s.lower()


@register('pprint')
def prettyprint(obj):
    """ Returns a pretty-printed representation of `obj`. """
    return pprint.pformat(obj)


@register
def pygmentize(text, lang=None):
    """ Applies syntax highlighting using Pygments.

    If no language is specified, Pygments will attempt to guess the correct
    lexer to use. If Pygments is not available or if an appropriate lexer
    cannot be found then the filter will return the input text with any
    html special characters escaped.
    """
    if pygments:
        if lang:
            try:
                lexer = pygments.lexers.get_lexer_by_name(lang)
            except pygments.util.ClassNotFound:
                lexer = None
        else:
            try:
                lexer = pygments.lexers.guess_lexer(text)
            except pygments.util.ClassNotFound:
                lexer = None
        if lexer:
            formatter = pygments.formatters.HtmlFormatter(nowrap=True)
            text = pygments.highlight(text, lexer, formatter)
        else:
            text = html.escape(text)
    else:
        text = html.escape(text)
    return text


@register
def random(seq):
    """ Returns a random element from the sequence `seq`. """
    return random.choice(seq)


@register('repr')
def to_repr(obj):
    """ Returns the result of calling repr() on `obj`. """
    return repr(obj)


@register
def slice(seq, start, stop=None, step=None):
    """ Returns the start:stop:step slice of the sequence `seq`. """
    return seq[start:stop:step]


@register
def spaceless(html):
    """ Strips all whitespace between html/xml tags. """
    return re.sub(r'>\s+<', '><', html)


@register
def startswith(s, prefix):
    """ True if the string starts with the specified prefix. """
    return s.startswith(prefix)


@register('str')
def to_str(obj):
    """ Returns the result of calling str() on `obj`. """
    return str(obj)


@register
def striptags(html):
    """ Returns the string `html` with all html tags stripped. """
    return re.sub(r'<[^>]*>', '', html)


@register
def teaser(s, delimiter='<!-- more -->'):
    """ Returns the portion of the string `s` before `delimiter`,
    or an empty string if `delimiter` is not found. """
    index = s.find(delimiter)
    if index == -1:
        return ''
    else:
        return s[:index]


@register
@register('title')
def titlecase(s):
    """ Returns the string `s` converted to titlecase. """
    return re.sub(
        r"[A-Za-z]+('[A-Za-z]+)?",
        lambda m: m.group(0)[0].upper() + m.group(0)[1:],
        s
    )


@register
def truncatechars(s, n, ellipsis='...'):
    """ Truncates the string `s` to at most `n` characters. """
    if len(s) > n:
        return s[:n - 3].rstrip(' .,;:?!') + ellipsis
    else:
        return s


@register
def truncatewords(s, n, ellipsis=' [...]'):
    """ Truncates the string `s` to at most `n` words. """
    words = s.split()
    if len(words) > n:
        return ' '.join(words[:n]) + ellipsis
    else:
        return ' '.join(words)


@register
def undefined(obj, fallback):
    """ Returns `obj` if `obj` is defined, otherwise `fallback`. """
    return fallback if isinstance(obj, errors.Undefined) else obj


@register
def upper(s):
    """ Returns the string `s` converted to uppercase. """
    return s.upper()


@register
def wrap(s, tag):
    """ Wraps a string in opening and closing tags. """
    return '<%s>%s</%s>' % (tag, str(s), tag)



# --------------------------------------------------------------------------
# The Template class is the public interface to the template engine.
#
# A Template object is initialized with a template string. It compiles the
# string and stores the resulting tree of template nodes. A compiled
# template can be rendered multiple times by calling its render() method
# with a dictionary of key-value pairs or a set of keyword arguments.
# --------------------------------------------------------------------------

import re
import collections

from . import config
from . import nodes

from .errors import Undefined, InvalidTag, NestingError


class Template:

    """ Public interface to the template engine. """

    def __init__(self, string):
        self.root = self._parse(Lexer(string))
        self.registry = self._register_blocks(self.root, {})

    def __repr__(self):
        return repr(self.root)

    def render(self, *pargs, **kwargs):
        """ Accepts a data dictionary or a set of keyword arguments. """
        data = pargs[0] if pargs else kwargs
        return self.root.render(Context(data, self))

    def _parse(self, tokens):
        """ Compiles a sequence of tokens into a tree of nodes.

        Syntax nodes with block scope remain on the stack until we
        encounter the corresponding end tag in the token stream. We
        then pop the stack and call .exit_scope() on the block node
        so it can process its child nodes.

        """
        root = nodes.nodemap['root']()
        stack, expecting = [root], []

        for token in tokens:
            if token.type == 'syntax':
                if token.tag in nodes.nodemap:
                    node = nodes.nodemap[token.tag](token)
                elif token.tag in nodes.nodemap['endtags']:
                    if not expecting:
                        raise NestingError('not expecting [%s]' % token.tag)
                    elif token.tag == expecting[-1]:
                        stack[-1].exit_scope()
                        stack.pop()
                        expecting.pop()
                        continue
                    else:
                        msg = 'expecting [%s], found [%s]'
                        raise NestingError(msg % (expecting[-1], token.tag))
                else:
                    msg = '[%s] is not a recognised template tag'
                    raise InvalidTag(msg % token.tag)
            else:
                node = nodes.nodemap[token.type](token)
            stack[-1].children.append(node)
            if node.end_tag:
                stack.append(node)
                expecting.append(node.end_tag)

        if expecting:
            raise NestingError('expecting [%s]' % expecting[-1])

        return stack.pop()

    def _register_blocks(self, node, registry):
        """ Assembles a registry of the template's {% block %} nodes.

        This function walks the node tree and assembles a dictionary
        containing lists of block nodes indexed by title. We use this list
        to implement template inheritance - a block node occurring later in
        a list overrides those occuring earlier.

        Note that we don't implement template inheritance by modifying the
        block nodes in situ. This is because templates can incorporate
        multiple (possibly-cached) sub-templates, so a single block node
        instance can form part of multiple distinct template trees.

        """
        if isinstance(node, nodes.nodemap['block']):
            registry.setdefault(node.title, []).append(node)
        for child in node.children:
            self._register_blocks(child, registry)
        return registry


class Context:

    """ Implements the template engine's variable lookup logic. """

    def __init__(self, data, template):

        # Data is stored on a stack of dictionaries.
        self.stack = []

        # Standard builtins.
        self.stack.append({'context': self, 'defined': self.defined})

        # User-configurable, per-application builtins.
        self.stack.append(config.builtins)

        # Instance-specific data.
        self.stack.append(data)

        # Nodes can store state information here to avoid threading issues.
        self.stash = {}

        # This reference gives nodes access to their parent template object.
        self.template = template

    def __setitem__(self, key, value):
        """ Assigns to the last dictionary on the stack. """
        self.stack[-1][key] = value

    def __getitem__(self, key):
        """ Checks every dictionary on the stack for `key`. """
        for d in reversed(self.stack):
            if key in d:
                return d[key]
        raise KeyError(key)

    def __delitem__(self, key):
        """ Deletes an item from the last dictionary on the stack. """
        del self.stack[-1][key]

    def __contains__(self, key):
        """ True if any dictionary on the stack contains `key`. """
        for d in self.stack:
            if key in d:
                return True
        return False

    def resolve(self, varstring):
        """ Attempts to resolve the object identified by `varstring`.

        If the variable name cannot be resolved an instance of the
        Undefined class is returned in its place.

        """
        obj = self
        for token in varstring.split('.'):
            try:
                obj = obj[token]
            except (TypeError, AttributeError, KeyError, ValueError):
                try:
                    obj = getattr(obj, token)
                except (TypeError, AttributeError):
                    obj = Undefined()
                    break
        return obj

    def push(self, data=None):
        """ Pushes a new data dictionary onto the stack. """
        self.stack.append(data or {})

    def pop(self):
        """ Pops the most recent data dictionary off the stack. """
        self.stack.pop()

    def defined(self, varstring):
        """ Returns true if `varstring` can be successfully resolved. """
        return not isinstance(self.resolve(varstring), Undefined)

    def get(self, key, default=None):
        """ Returns `default` if `key` cannot be found on the stack. """
        for d in reversed(self.stack):
            if key in d:
                return d[key]
        return default

    def update(self, data):
        """ Updates the most recent dictionary on the data stack. """
        self.stack[-1].update(data)


class Lexer:

    """ Transforms an input string into a sequence of tokens. """

    token = collections.namedtuple('Token', 'type tag content')

    def __init__(self, string):
        self.string = string
        self.regex = re.compile(r'(%s.*?%s|%s.*?%s|%s.*?%s|%s.*?%s)' % (
            re.escape(config.delimiters['syntax_start']),
            re.escape(config.delimiters['syntax_end']),
            re.escape(config.delimiters['eprint_start']),
            re.escape(config.delimiters['eprint_end']),
            re.escape(config.delimiters['print_start']),
            re.escape(config.delimiters['print_end']),
            re.escape(config.delimiters['comment_start']),
            re.escape(config.delimiters['comment_end']),
        ), re.DOTALL)

    def __iter__(self):
        for fragment in self.regex.split(self.string):
            if fragment:
                token = self._get_token(fragment)
                if token.content:
                    yield token

    def _get_token(self, fragment):
        if fragment.startswith(config.delimiters['eprint_start']):
            content = self._strip(fragment, 'eprint_start', 'eprint_end')
            type = 'eprint'
            tag = None

        elif fragment.startswith(config.delimiters['print_start']):
            content = self._strip(fragment, 'print_start', 'print_end')
            type = 'print'
            tag = None

        elif fragment.startswith(config.delimiters['syntax_start']):
            content = self._strip(fragment, 'syntax_start', 'syntax_end')
            type = 'syntax'
            tag = content.split(' ')[0]

        elif fragment.startswith(config.delimiters['comment_start']):
            content = None
            type = 'comment'
            tag = None

        else:
            content = fragment
            type = 'text'
            tag = None

        return self.token(type, tag, content)

    def _strip(self, fragment, start, end):
        len_start = len(config.delimiters[start])
        len_end = len(config.delimiters[end])
        return fragment[len_start:-len_end].strip()



#!/usr/bin/env python3
"""
Ibis
====

Features:

* Django/Jinja-style syntax.
* Supports looping, conditionals, filters, template inheritance.
* Extensible. Easily add custom template tags and filters.
* Self-contained, no dependencies. Use as a drop-in component in any project.
* Public domain code. No license compatibility issues.

Sample syntax::

    <ul>
        {% for post in posts %}
            <li><a href="{{ post.url }}">{{ post.title }}</a></li>
        {% endfor %}
    </ul>

See the `package documentation <http://mulholland.xyz/docs/ibis/>`_ or the
project's `Github homepage <https://github.com/dmulholland/ibis>`_ for
further details.

"""

import os
import re
import io

from setuptools import setup, find_packages


filepath = os.path.join(os.path.dirname(__file__), 'ibis', '__init__.py')
with io.open(filepath, encoding='utf-8') as metafile:
    regex = r'''^__([a-z]+)__ = ["'](.*)["']'''
    meta = dict(re.findall(regex, metafile.read(), flags=re.MULTILINE))


setup(
    name = 'ibis',
    version = meta['version'],
    packages =  find_packages(),
    include_package_data = True,
    author = 'Darren Mulholland',
    url='https://github.com/dmulholland/ibis',
    license = 'Public Domain',
    description = (
        'A lightweight template engine.'
    ),
    long_description = __doc__,
    classifiers = [
        'Programming Language :: Python :: 3',
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'License :: Public Domain',
        'Intended Audience :: Developers',
        'Topic :: Text Processing :: General',
        'Topic :: Text Processing :: Markup :: HTML',
    ],
)
